/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Reads .unv file as written by Salome.
   
    Not extensively tested.

\*---------------------------------------------------------------------------*/

#include "argList.H"
#include "Time.H"
#include "polyMesh.H"
#include "IFstream.H"
#include "polyPatch.H"
#include "ListOps.H"
#include "cellModeller.H"

#include <fstream>

using namespace Foam;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

// Element type numbers
static label UNVBEAM = 11;
static label UNVTRI = 41;
static label UNVTET = 111;

static label starttag = -1;
static label endtag = -1;

static label nodetag = 2411;
static label elemtag = 2412;

void renumber
(
    const Map<label>& mshToFoam,
    labelList& labels
)
{
    forAll(labels, labelI)
    {
        labels[labelI] = mshToFoam[labels[labelI]];
    }
}

// Main program:

int main(int argc, char *argv[])
{
    argList::noParallel();
    argList::validArgs.append(".unv file");

#   include "setRootCase.H"
#   include "createTime.H"

    fileName mshName(args.args()[3]);

    IFstream inFile1(mshName);
    IFstream inFile2(mshName);

    label tag;

    label dk;

    label nIndex;
    scalar xVal, yVal, zVal;

    labelList beamPoints(2);
    labelList triPoints(3);
    labelList tetPoints(4);

    label eIndex;
    label eType;
    label nNodes;

    label nBeam = 0;
    label nTri = 0;
    label nTet = 0;

    label nPoints = 0;
    label nCells = 0;

    label a;

    a = 0;

    do
      {

	inFile1 >> tag;

	if (a == 2) break;

	//Info << "Tag:" << tag << endl;
	
	if (tag == starttag)
	  {

	    inFile1 >> tag;
		
	    if (tag == nodetag)
	      {

		a++;

		do
		  {
		    
		    inFile1 >> tag;

		    //Info << "Tag:" << tag << endl;
		
		    if (tag != endtag) 
		      {
			nIndex = tag;

			//Info << "Node index:" << nIndex << endl;

			inFile1 >> dk >> dk >> dk;	
			inFile1 >> xVal >> yVal >> zVal;		

			//Info << xVal << " " << yVal << " " << zVal << endl;

			nPoints++;

		      }
		    else
		      break;


		  } while (!inFile1.eof());
	      }

	    if (tag == elemtag)
	      {

		a++;

		do
		  {
		
		    inFile1 >> tag;

		    //Info << "Tag:" << tag << endl;
		    
		    if (tag != endtag) 
		      {
			eIndex = tag;

			//Info << "Element index:" << eIndex << endl;

			inFile1 >> eType >> dk >> dk >> dk >> nNodes;	

			if (eType == UNVBEAM)
			  {
			    inFile1 >> dk >> dk >> dk;			   			   
			    inFile1 >> beamPoints[0] >> beamPoints[1];

			    //Info << beamPoints[0] << " " << beamPoints[1] << endl;
			    
			    nBeam++;
			  }

			if (eType == UNVTRI)
			  {
			    inFile1 >> triPoints[0] >> triPoints[1] >> triPoints[2];

			    //Info << triPoints[0] << " " << triPoints[1] << " " << triPoints[2] << endl;

			    nTri++;
			  }
			
			if (eType == UNVTET)
			  {
			    inFile1 >> tetPoints[0] >> tetPoints[1] >> tetPoints[2] >> tetPoints[3];

			    //Info << tetPoints[0] << " " << tetPoints[1] <<  " " << tetPoints[2] << " " << tetPoints[3] << endl; 
			    
			    nTet++;

			    nCells++;

			  }

		      }
		    else
		      break;

		  } while (!inFile1.eof());
	      }
	    
	  }

      } while (!inFile1.eof());

    Info<< "Number of points: " << nPoints << endl;
    Info<< "Number of cells: " << nCells << endl;

    label pointI = 0;
    label cellI = 0;
 
    const cellModel& tet = *(cellModeller::lookup("tet"));

    pointField points(nPoints);
    cellShapeList cells(nCells);

    Map<label> mshToFoam(nPoints);

    a = 0;

    do
      {

	inFile2 >> tag;

	if (a == 2) break;

	//Info << "Tag:" << tag << endl;
	
	if (tag == starttag)
	  {

	    inFile2 >> tag;
		
	    if (tag == nodetag)
	      {

		a++;

		do
		  {
		    
		    inFile2 >> tag;

		    //Info << "Tag:" << tag << endl;
		
		    if (tag != endtag) 
		      {
			nIndex = tag;

			//Info << "Node index:" << nIndex << endl;

			inFile2 >> dk >> dk >> dk;	
			inFile2 >> xVal >> yVal >> zVal;		

			//Info << xVal << " " << yVal << " " << zVal << endl;
			
			point& pt = points[pointI];

			pt.x() = xVal;
			pt.y() = yVal;
			pt.z() = zVal;

			mshToFoam.insert(nIndex, pointI);

			pointI++;

		      }
		    else
		      break;


		  } while (!inFile2.eof());
	      }

	    if (tag == elemtag)
	      {

		a++;

		do
		  {
		
		    inFile2 >> tag;

		    //Info << "Tag:" << tag << endl;
		    
		    if (tag != endtag) 
		      {
			eIndex = tag;

			//Info << "Element index:" << eIndex << endl;

			inFile2 >> eType >> dk >> dk >> dk >> nNodes;	

			if (eType == UNVBEAM)
			  {
			    inFile2 >> dk >> dk >> dk;			   			   
			    inFile2 >> beamPoints[0] >> beamPoints[1];

			    //Info << beamPoints[0] << " " << beamPoints[1] << endl;
			    
			    nBeam++;
			  }

			if (eType == UNVTRI)
			  {
			    inFile2 >> triPoints[0] >> triPoints[1] >> triPoints[2];

			    //Info << triPoints[0] << " " << triPoints[1] << " " << triPoints[2] << endl;

			    nTri++;
			  }
			
			if (eType == UNVTET)
			  {
			    inFile2 >> tetPoints[0] >> tetPoints[1] >> tetPoints[2] >> tetPoints[3];

			    //Info << tetPoints[0] << " " << tetPoints[1] <<  " " << tetPoints[2] << " " << tetPoints[3] << endl; 

			    nTet++;

			    renumber(mshToFoam, tetPoints);			    

			    cells[cellI++] = cellShape(tet, tetPoints);

			    nCells++;

			  }

		      }
		    else
		      break;

		  } while (!inFile2.eof());
	      }
	    
	  }

      } while (!inFile2.eof());

    polyMesh mesh
      (
        IOobject
        (
            polyMesh::defaultRegion,
            runTime.constant(),
            runTime
        ),
        points,
        cells,
        faceListList(0),
        wordList(0),
        wordList(0),
        polyPatch::typeName,
        wordList(0)
    );
   
    //Get polyMesh to write to constant
    runTime.setTime(instant(runTime.constant()), 0);

    //Write mesh
    Info << "Writing mesh ..." << endl;

    mesh.write();
    
    Info << "End\n" << endl;

    return 0;
}


// ************************************************************************* //

